//
//  HUDNode.m
//  SuperKoalio
//
//  Created by Jake Gundersen on 3/1/14.
//  Copyright (c) 2014 Razeware. All rights reserved.
//

#import "HUDNode.h"
#import "Joystick.h"

@import GameController;

NSString * const kGameTogglePauseNotification = @"GameTogglePauseNotification";

@interface HUDNode()

@property (nonatomic, strong) Joystick *joystick;

@property (nonatomic, strong) SKSpriteNode *aButton;
@property (nonatomic, strong) SKSpriteNode *bButton;

@property (nonatomic, strong) UITouch *dashTouch;
@property (nonatomic, strong) GCController *controller;
@property (nonatomic, strong) id connectObserver;
@property (nonatomic, strong) id disconnectObserver;
@property (nonatomic, strong) NSMutableArray *snapShots;
@property (nonatomic, assign) NSUInteger currentSnapshotIndex;
@property (nonatomic, assign) BOOL shouldReplaySnapshot;

@end

@implementation HUDNode

- (instancetype)initWithSize:(CGSize)size {
    if ((self = [super init])) {
        //2
        self.userInteractionEnabled = YES;
        //3
        SKSpriteNode *joyBack = [SKSpriteNode spriteNodeWithImageNamed:@"dpad"];
        SKSpriteNode *joyFront = [SKSpriteNode spriteNodeWithImageNamed:@"joystick"];
        joyBack.alpha = 0.5;
        joyFront.alpha = 0.7;
        //4
        self.joystick = [Joystick joystickWithThumb:joyFront andBackdrop:joyBack];
        //5
        self.joystick.position = CGPointMake(90, 160);
        [self addChild:self.joystick];
        //6
        self.aButton = [SKSpriteNode spriteNodeWithImageNamed:@"buttonA"];
        self.aButton.position = CGPointMake(size.width - 40, 120);
        [self addChild:self.aButton];
        
        self.bButton = [SKSpriteNode spriteNodeWithImageNamed:@"buttonB"];
        self.bButton.position = CGPointMake(size.width - 140, 120);
        [self addChild:self.bButton];
        
        //Add hardware controller code here
        if ([[GCController controllers] count]) {
          [self toggleHardwareController:YES];
        }
      
        //Add observers here
        __weak typeof(self) weakself = self;
        self.connectObserver = [[NSNotificationCenter defaultCenter] addObserverForName:GCControllerDidConnectNotification object:nil queue:[NSOperationQueue mainQueue] usingBlock:^(NSNotification *note) {
          if ([[GCController controllers] count] == 1) {
            [weakself toggleHardwareController:YES];
          }
        }];
        self.disconnectObserver = [[NSNotificationCenter defaultCenter] addObserverForName:GCControllerDidDisconnectNotification object:nil queue:[NSOperationQueue mainQueue] usingBlock:^(NSNotification *note) {
          if (![[GCController controllers] count]) {
            [weakself toggleHardwareController:NO];
          }
        }];
      
        //[self recordSnapshots];
        //[self replaySnapshots];
    }
    return self;
}

- (void)dealloc
{
  [[NSNotificationCenter defaultCenter] removeObserver:self.connectObserver];
  [[NSNotificationCenter defaultCenter] removeObserver:self.disconnectObserver];
}

- (void)toggleHardwareController:(BOOL)useHardware {
  //1
  if (useHardware) {
    //2
    self.alpha = 0.0;
    //3
    self.controller = [GCController controllers][0];
    self.controller.playerIndex = 0;
    //4
    if (self.controller.gamepad) {
      //5
      __weak typeof(self) weakself = self;
      [self.controller.gamepad.buttonA setValueChangedHandler:^(GCControllerButtonInput *button, float value, BOOL pressed) {
        weakself.aPushed = pressed;
      }];
      [self.controller.gamepad.buttonX setValueChangedHandler:^(GCControllerButtonInput *button, float value, BOOL pressed) {
        weakself.shouldDash = pressed;
      }];
      //Add controller pause handler here
      [self.controller setControllerPausedHandler:^(GCController *controller) {
        [[NSNotificationCenter defaultCenter] postNotificationName:kGameTogglePauseNotification object:nil];
      }];
    }
  } else {
    //7
    self.alpha = 1.0;
    self.controller = nil;
  }
}

- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event {
    for (UITouch *touch in touches) {
        //1
        CGPoint touchLocation = [touch locationInNode:self];
        //2
        if (CGRectContainsPoint(self.bButton.frame, touchLocation)) {
            self.shouldDash = YES;
            self.dashTouch = touch;
            self.bButton.alpha = 0.7;
        }
        //3
        if (CGRectContainsPoint(self.aButton.frame, touchLocation)) {
            self.aPushed = YES;
            self.aButton.alpha = 0.7;
        }
    }
}

- (void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event {
    for (UITouch *touch in touches) {
        //1
        CGPoint touchLocation = [touch locationInNode:self];
        CGPoint previousTouchLocation = [touch previousLocationInNode:self];
        //2
        if (CGRectContainsPoint(self.aButton.frame, previousTouchLocation) &&
            !CGRectContainsPoint(self.aButton.frame, touchLocation)) {
            self.aPushed = NO;
            self.aButton.alpha = 1.0;
        }
        //3
        if (!CGRectContainsPoint(self.aButton.frame, previousTouchLocation) &&
            CGRectContainsPoint(self.aButton.frame, touchLocation)) {
            self.aPushed = YES;
            self.aButton.alpha = 0.7;
        }
    }
}

- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event {
    for (UITouch *touch in touches) {
        //1
        if (self.dashTouch == touch) {
            self.shouldDash = NO;
            self.dashTouch = nil;
            self.bButton.alpha = 1.0;
        }
        //2
        CGPoint touchLocation = [touch locationInNode:self];
        if (CGRectContainsPoint(self.aButton.frame, touchLocation)) {
            self.aPushed = NO;
            self.aButton.alpha = 1.0;
        }
    }
}

- (float)xJoystickVelocity {

  //1
  if (self.shouldReplaySnapshot) {
    //2
    id currentSnapshot = [self currentGamepadFromSnapshot];
    //3
    self.currentSnapshotIndex++;
    //4
    if ([currentSnapshot isKindOfClass:[GCGamepadSnapshot class]]) {
      //5
      GCGamepadSnapshot *gamepadSnapshot = (GCGamepadSnapshot *)currentSnapshot;
      //6
      return gamepadSnapshot.dpad.xAxis.value;
    } else {
      //7
      GCExtendedGamepadSnapshot *extendedGamepadSnapshot = (GCExtendedGamepadSnapshot *)currentSnapshot;
      return extendedGamepadSnapshot.leftThumbstick.xAxis.value;
    }
  }

  if (self.controller.extendedGamepad) {
    //1
    if (self.shouldRecordSnapshots) {
      //2
      NSData *snapShot = [[self.controller.extendedGamepad saveSnapshot] snapshotData];
      //3
      NSDictionary *snapshotDict = @{@"type": @"extended", @"data":snapShot};
      //4
      [self.snapShots addObject:snapshotDict];
   
    }
    return self.controller.extendedGamepad.leftThumbstick.xAxis.value;
  } else if (self.controller.gamepad) {
    //5
    if (self.shouldRecordSnapshots) {
      NSData *snapShot = [[self.controller.gamepad saveSnapshot] snapshotData];
      NSDictionary *snapshotDict = @{@"type": @"gamepad", @"data":snapShot};
      [self.snapShots addObject:snapshotDict];
    }
    return self.controller.gamepad.dpad.xAxis.value;
  }
  //3
  return self.joystick.velocity.x / 60.0;
}

- (NSURL *)snapShotDataPath {
  //1
  NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
  NSString *filePath = paths[0];
  filePath = [filePath stringByAppendingPathComponent:@"snapshotData.plist"];
  //2
  return [NSURL fileURLWithPath:filePath];
}
 
- (void)saveSnapshotsToDisk:(NSArray *)snapShots {
  //3
  if (![snapShots count]) return;
  //4
  if (![snapShots writeToURL:[self snapShotDataPath] atomically:YES]) {
    NSLog(@"Couldn't save snapshots array to file");
  }
}

- (BOOL)recordSnapshots {
  if (!self.controller || self.shouldRecordSnapshots) return NO;
  self.shouldRecordSnapshots = YES;
  self.snapShots = [NSMutableArray array];
  return YES;
}

- (void)saveSnapshots {
  [self saveSnapshotsToDisk:self.snapShots];
}

- (BOOL)replaySnapshots {
  //1
  self.snapShots = [NSArray arrayWithContentsOfURL:[self snapShotDataPath]];
  //2
  if (!self.snapShots || ![self.snapShots count]) return NO;
  //3
  self.shouldReplaySnapshot = YES;
  //4
  return YES;
}

- (id)currentGamepadFromSnapshot {
  //1
  NSDictionary *currentSnapshotData = self.snapShots[self.currentSnapshotIndex];
  //2
  id snapshot = nil;
  //3
  if ([currentSnapshotData[@"type"] isEqualToString:@"gamepad"]) {
    //4
    snapshot = [[GCGamepadSnapshot alloc] initWithSnapshotData:currentSnapshotData[@"data"]];
  } else {
    //5
    snapshot = [[GCExtendedGamepadSnapshot alloc] initWithSnapshotData:currentSnapshotData[@"data"]];
  }
  return snapshot;
}

- (BOOL)shouldDash {
  //1
  if (self.shouldReplaySnapshot) {
    //2
    id snapshot = [self currentGamepadFromSnapshot];
    //3
    return [[snapshot valueForKeyPath:@"buttonX.pressed"] boolValue];
  }
  //4
  return _shouldDash;
}

- (BOOL)aPushed {
  if (self.shouldReplaySnapshot) {
    id snapshot = [self currentGamepadFromSnapshot];
    return [[snapshot valueForKeyPath:@"buttonA.pressed"] boolValue];
  }
  return _aPushed;
}

@end
